@@ -39,7 +39,7 @@
 #include <dlfcn.h>
 #include <unistd.h>
 #ifdef ANDROID
-#include <cutils/log.h>
+#include <log/log.h>
 /* support versions < JellyBean */
 #ifndef ALOGE
 #define ALOGE LOGE
@@ -52,7 +52,7 @@
 #define DRIVER_EXTENSION	"_drv_video.so"
 
 #define ASSERT		assert
-#define CHECK_VTABLE(s, ctx, func) if (!va_checkVtable(dpy, ctx->vtable->va##func, #func)) s = VA_STATUS_ERROR_UNKNOWN;
+#define CHECK_VTABLE(s, ctx, func) if (!va_checkVtable(dpy, ctx->vtable->va##func, #func)) s = VA_STATUS_ERROR_UNIMPLEMENTED;
 #define CHECK_MAXIMUM(s, ctx, var) if (!va_checkMaximum(dpy, ctx->max_##var, #var)) s = VA_STATUS_ERROR_UNKNOWN;
 #define CHECK_STRING(s, ctx, var) if (!va_checkString(dpy, ctx->str_##var, #var)) s = VA_STATUS_ERROR_UNKNOWN;
 
@@ -341,12 +341,82 @@ va_getDriverInitName(char *name, int namelen, int major, int minor)
     int ret = snprintf(name, namelen, "__vaDriverInit_%d_%d", major, minor);
     return ret > 0 && ret < namelen;
 }
-
-static VAStatus va_getDriverName(VADisplay dpy, char **driver_name)
+/** retrieve the back end driver candidate num , by default it should be 1
+  * if there are no vaGetNumCandidates implementation in the display context
+  * it should be 1 to avoid backward compatible issue */
+static VAStatus va_getDriverNumCandidates(VADisplay dpy, int *num_candidates)
 {
     VADisplayContextP pDisplayContext = (VADisplayContextP)dpy;
+    *num_candidates = 1;
+    const char *driver_name_env = NULL;
+    VAStatus vaStatus = VA_STATUS_SUCCESS;
+    VADriverContextP ctx;
 
-    return pDisplayContext->vaGetDriverName(pDisplayContext, driver_name);
+    ctx = CTX(dpy);
+    driver_name_env = getenv("LIBVA_DRIVER_NAME");
+
+    if (pDisplayContext->vaGetNumCandidates)
+        vaStatus = pDisplayContext->vaGetNumCandidates(pDisplayContext, num_candidates);
+    if ((ctx->override_driver_name) || (driver_name_env && (geteuid() == getuid())))
+        *num_candidates = 1;
+    return vaStatus;
+}
+
+static VAStatus va_getDriverNameByIndex(VADisplay dpy, char **driver_name, int candidate_index)
+{
+    VADisplayContextP pDisplayContext = (VADisplayContextP)dpy;
+    const char *driver_name_env = NULL;
+    VADriverContextP ctx;
+    VAStatus status = VA_STATUS_SUCCESS;
+
+    ctx = CTX(dpy);
+    if (pDisplayContext->vaGetDriverNameByIndex) {
+        /*if vaGetDriverNameByIndex is implemented*/
+        status = pDisplayContext->vaGetDriverNameByIndex(pDisplayContext, driver_name, candidate_index);
+    } else {
+        if (candidate_index == 0)
+            status = pDisplayContext->vaGetDriverName(pDisplayContext, driver_name);
+        else
+            status = VA_STATUS_ERROR_INVALID_PARAMETER;
+    }
+    driver_name_env = getenv("LIBVA_DRIVER_NAME");
+    /*if user set driver name by vaSetDriverName */
+    if (ctx->override_driver_name){
+        if(*driver_name)
+            free(*driver_name);
+        *driver_name = strdup(ctx->override_driver_name);
+        if (!(*driver_name)) {
+              va_errorMessage(dpy, "va_getDriverNameByIndex  failed with %s, out of memory\n",vaErrorStr(VA_STATUS_ERROR_ALLOCATION_FAILED));
+            return VA_STATUS_ERROR_ALLOCATION_FAILED;
+        }
+        va_infoMessage(dpy, "User requested driver '%s'\n", *driver_name);
+        return VA_STATUS_SUCCESS;
+    } else if (driver_name_env && (geteuid() == getuid())) {
+        if(*driver_name)
+            free(*driver_name);
+        /*if user set driver name by environment variable*/
+        *driver_name = strdup(driver_name_env);
+        va_infoMessage(dpy, "User environment variable requested driver '%s'\n", *driver_name);
+        return VA_STATUS_SUCCESS;
+    }
+    return status;
+}
+
+static char *va_getDriverPath(const char *driver_dir, const char *driver_name)
+{
+  int n = snprintf(0, 0, "%s/%s%s", driver_dir, driver_name, DRIVER_EXTENSION);
+  if (n < 0)
+      return NULL;
+  char *driver_path = (char *) malloc(n + 1);
+  if (!driver_path)
+      return NULL;
+  n = snprintf(driver_path, n + 1, "%s/%s%s",
+               driver_dir, driver_name, DRIVER_EXTENSION);
+  if (n < 0) {
+    free(driver_path);
+    return NULL;
+  }
+  return driver_path;
 }
 
 static VAStatus va_openDriver(VADisplay dpy, char *driver_name)
@@ -364,24 +434,22 @@ static VAStatus va_openDriver(VADisplay dpy, char *driver_name)
         search_path = VA_DRIVERS_PATH;
 
     search_path = strdup((const char *)search_path);
+    if (!search_path) {
+         va_errorMessage(dpy, "%s L%d Out of memory\n",
+                            __FUNCTION__, __LINE__);
+         return VA_STATUS_ERROR_ALLOCATION_FAILED;
+    }
     driver_dir = strtok_r(search_path, ":", &saveptr);
     while (driver_dir) {
         void *handle = NULL;
-        char *driver_path = (char *) malloc( strlen(driver_dir) +
-                                             strlen(driver_name) +
-                                             strlen(DRIVER_EXTENSION) + 2 );
+        char *driver_path = va_getDriverPath(driver_dir, driver_name);
         if (!driver_path) {
-            va_errorMessage(dpy, "%s L%d Out of memory!n",
+            va_errorMessage(dpy, "%s L%d Out of memory\n",
                             __FUNCTION__, __LINE__);
             free(search_path);
             return VA_STATUS_ERROR_ALLOCATION_FAILED;
         }
 
-        strncpy( driver_path, driver_dir, strlen(driver_dir) + 1);
-        strncat( driver_path, "/", strlen("/") );
-        strncat( driver_path, driver_name, strlen(driver_name) );
-        strncat( driver_path, DRIVER_EXTENSION, strlen(DRIVER_EXTENSION) );
-        
         va_infoMessage(dpy, "Trying to open %s\n", driver_path);
 #ifndef ANDROID
         handle = dlopen( driver_path, RTLD_NOW | RTLD_GLOBAL | RTLD_NODELETE );
@@ -397,15 +465,16 @@ static VAStatus va_openDriver(VADisplay dpy, char *driver_name)
             char init_func_s[256];
             int i;
 
-            static const struct {
+            struct {
                 int major;
                 int minor;
-            } compatible_versions[] = {
-                { VA_MAJOR_VERSION, VA_MINOR_VERSION },
-                { VA_MAJOR_VERSION, 1 },
-                { VA_MAJOR_VERSION, 0 },
-                { -1, }
-            };
+            } compatible_versions[VA_MINOR_VERSION + 2];
+            for (i = 0; i <= VA_MINOR_VERSION; i ++) {
+                compatible_versions[i].major = VA_MAJOR_VERSION;
+                compatible_versions[i].minor = VA_MINOR_VERSION - i;
+            }
+            compatible_versions[i].major = -1;
+            compatible_versions[i].minor = -1;
 
             for (i = 0; compatible_versions[i].major >= 0; i++) {
                 if (va_getDriverInitName(init_func_s, sizeof(init_func_s),
@@ -475,7 +544,6 @@ static VAStatus va_openDriver(VADisplay dpy, char *driver_name)
                     CHECK_VTABLE(vaStatus, ctx, EndPicture);
                     CHECK_VTABLE(vaStatus, ctx, SyncSurface);
                     CHECK_VTABLE(vaStatus, ctx, QuerySurfaceStatus);
-                    CHECK_VTABLE(vaStatus, ctx, PutSurface);
                     CHECK_VTABLE(vaStatus, ctx, QueryImageFormats);
                     CHECK_VTABLE(vaStatus, ctx, CreateImage);
                     CHECK_VTABLE(vaStatus, ctx, DeriveImage);
@@ -603,13 +671,6 @@ const char *vaErrorStr(VAStatus error_status)
     return "unknown libva error / description missing";
 }
 
-const static char *prefer_driver_list[4] = {
-    "i965",
-    "hybrid",
-    "pvr",
-    "iHD",
-};
-
 VAStatus vaSetDriverName(
     VADisplay dpy,
     char *driver_name
@@ -618,15 +679,8 @@ VAStatus vaSetDriverName(
     VADriverContextP ctx;
     VAStatus vaStatus = VA_STATUS_SUCCESS;
     char *override_driver_name = NULL;
-    int i, found;
     ctx = CTX(dpy);
 
-    if (geteuid() != getuid()) {
-        vaStatus = VA_STATUS_ERROR_OPERATION_FAILED;
-        va_errorMessage(dpy, "no permission to vaSetDriverName\n");
-        return vaStatus;
-    }
-
     if (strlen(driver_name) == 0 || strlen(driver_name) >=256) {
         vaStatus = VA_STATUS_ERROR_INVALID_PARAMETER;
         va_errorMessage(dpy, "vaSetDriverName returns %s\n",
@@ -634,25 +688,7 @@ VAStatus vaSetDriverName(
         return vaStatus;
     }
 
-    found = 0;
-    for (i = 0; i < sizeof(prefer_driver_list) / sizeof(char *); i++) {
-        if (strlen(prefer_driver_list[i]) != strlen(driver_name))
-            continue;
-        if (!strncmp(prefer_driver_list[i], driver_name, strlen(driver_name))) {
-            found = 1;
-            break;
-        }
-    }
-
-    if (!found) {
-        vaStatus = VA_STATUS_ERROR_INVALID_PARAMETER;
-        va_errorMessage(dpy, "vaSetDriverName returns %s. Incorrect parameter\n",
-                         vaErrorStr(vaStatus));
-        return vaStatus;
-    }
-
     override_driver_name = strdup(driver_name);
-
     if (!override_driver_name) {
         vaStatus = VA_STATUS_ERROR_ALLOCATION_FAILED;
         va_errorMessage(dpy, "vaSetDriverName returns %s. Out of Memory\n",
@@ -670,15 +706,13 @@ VAStatus vaInitialize (
     int *minor_version 	 /* out */
 )
 {
-    const char *driver_name_env = NULL;
     char *driver_name = NULL;
+    int  num_candidates = 1;
+    int  candidate_index = 0;
     VAStatus vaStatus;
-    VADriverContextP ctx;
 
     CHECK_DISPLAY(dpy);
 
-    ctx = CTX(dpy);
-
     va_TraceInit(dpy);
 
     va_FoolInit(dpy);
@@ -686,52 +720,37 @@ VAStatus vaInitialize (
     va_MessagingInit();
 
     va_infoMessage(dpy, "VA-API version %s\n", VA_VERSION_S);
-
-    vaStatus = va_getDriverName(dpy, &driver_name);
-
-    if (!ctx->override_driver_name) {
-        va_infoMessage(dpy, "va_getDriverName() returns %d\n", vaStatus);
-
-        driver_name_env = getenv("LIBVA_DRIVER_NAME");
-    } else if (vaStatus == VA_STATUS_SUCCESS) {
-        if (driver_name)
+    /*get backend driver candidate number, by default the value should be 1*/
+    vaStatus = va_getDriverNumCandidates(dpy, &num_candidates);
+    if (vaStatus != VA_STATUS_SUCCESS) {
+        num_candidates = 1;
+    }
+    /*load driver one by one, until load success */
+    for (candidate_index = 0; candidate_index < num_candidates; candidate_index ++) {
+        if(driver_name)
             free(driver_name);
-
-        driver_name = strdup(ctx->override_driver_name);
-        if (!driver_name) {
-            vaStatus = VA_STATUS_ERROR_ALLOCATION_FAILED;
-            va_errorMessage(dpy, "vaInitialize() failed with %s, out of memory\n",
-                        vaErrorStr(vaStatus));
-            return vaStatus;
+        vaStatus = va_getDriverNameByIndex(dpy, &driver_name, candidate_index);
+        if(vaStatus != VA_STATUS_SUCCESS) {
+            va_errorMessage(dpy, "vaGetDriverNameByIndex() failed with %s, driver_name = %s\n", vaErrorStr(vaStatus), driver_name);
+            break;
         }
-        va_infoMessage(dpy, "User requested driver '%s'\n", driver_name);
-    }
-
-    if (driver_name_env && (geteuid() == getuid())) {
-        /* Don't allow setuid apps to use LIBVA_DRIVER_NAME */
-        if (driver_name) /* memory is allocated in va_getDriverName */
-            free(driver_name);
-        
-        driver_name = strdup(driver_name_env);
-        vaStatus = VA_STATUS_SUCCESS;
-        va_infoMessage(dpy, "User requested driver '%s'\n", driver_name);
-    }
-
-    if ((VA_STATUS_SUCCESS == vaStatus) && (driver_name != NULL)) {
         vaStatus = va_openDriver(dpy, driver_name);
         va_infoMessage(dpy, "va_openDriver() returns %d\n", vaStatus);
 
-        *major_version = VA_MAJOR_VERSION;
-        *minor_version = VA_MINOR_VERSION;
-    } else
-        va_errorMessage(dpy, "va_getDriverName() failed with %s,driver_name=%s\n",
-                        vaErrorStr(vaStatus), driver_name);
+        if (vaStatus == VA_STATUS_SUCCESS) {
+            break;
+       }
+
+    }
+
+    *major_version = VA_MAJOR_VERSION;
+    *minor_version = VA_MINOR_VERSION;
 
     if (driver_name)
         free(driver_name);
     
     VA_TRACE_LOG(va_TraceInitialize, dpy, major_version, minor_version);
-
+    VA_TRACE_RET(dpy, vaStatus);
     return vaStatus;
 }
 
@@ -766,6 +785,7 @@ VAStatus vaTerminate (
   }
 
   VA_TRACE_LOG(va_TraceTerminate, dpy);
+  VA_TRACE_RET(dpy, vaStatus);
 
   va_TraceEnd(dpy);
 
@@ -838,10 +858,13 @@ VAStatus vaQueryConfigEntrypoints (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaQueryConfigEntrypoints ( ctx, profile, entrypoints, num_entrypoints);
+  vaStatus = ctx->vtable->vaQueryConfigEntrypoints ( ctx, profile, entrypoints, num_entrypoints);
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 VAStatus vaGetConfigAttributes (
@@ -853,10 +876,13 @@ VAStatus vaGetConfigAttributes (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaGetConfigAttributes ( ctx, profile, entrypoint, attrib_list, num_attribs );
+  vaStatus = ctx->vtable->vaGetConfigAttributes ( ctx, profile, entrypoint, attrib_list, num_attribs );
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 VAStatus vaQueryConfigProfiles (
@@ -866,10 +892,13 @@ VAStatus vaQueryConfigProfiles (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaQueryConfigProfiles ( ctx, profile_list, num_profiles );
+  vaStatus =  ctx->vtable->vaQueryConfigProfiles ( ctx, profile_list, num_profiles );
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 VAStatus vaCreateConfig (
@@ -892,7 +921,7 @@ VAStatus vaCreateConfig (
   /* record the current entrypoint for further trace/fool determination */
   VA_TRACE_ALL(va_TraceCreateConfig, dpy, profile, entrypoint, attrib_list, num_attribs, config_id);
   VA_FOOL_FUNC(va_FoolCreateConfig, dpy, profile, entrypoint, attrib_list, num_attribs, config_id);
-  
+  VA_TRACE_RET(dpy, vaStatus); 
   return vaStatus;
 }
 
@@ -910,6 +939,7 @@ VAStatus vaDestroyConfig (
   vaStatus = ctx->vtable->vaDestroyConfig ( ctx, config_id );
 
   VA_TRACE_ALL(va_TraceDestroyConfig, dpy, config_id);
+  VA_TRACE_RET(dpy, vaStatus);
 
   return vaStatus;
 }
@@ -924,10 +954,13 @@ VAStatus vaQueryConfigAttributes (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaQueryConfigAttributes( ctx, config_id, profile, entrypoint, attrib_list, num_attribs);
+  vaStatus = ctx->vtable->vaQueryConfigAttributes( ctx, config_id, profile, entrypoint, attrib_list, num_attribs);
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 VAStatus vaQueryProcessingRate (
@@ -938,11 +971,15 @@ VAStatus vaQueryProcessingRate (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
   if(!ctx->vtable->vaQueryProcessingRate)
-      return VA_STATUS_ERROR_UNIMPLEMENTED;
-  return ctx->vtable->vaQueryProcessingRate( ctx, config_id, proc_buf, processing_rate);
+      vaStatus = VA_STATUS_ERROR_UNIMPLEMENTED;
+  else
+      vaStatus = ctx->vtable->vaQueryProcessingRate( ctx, config_id, proc_buf, processing_rate);
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 /* XXX: this is a slow implementation that will be removed */
@@ -973,7 +1010,7 @@ va_impl_query_surface_attributes(
         { VASurfaceAttribMinHeight,     VAGenericValueTypeInteger },
         { VASurfaceAttribMaxHeight,     VAGenericValueTypeInteger },
         { VASurfaceAttribMemoryType,    VAGenericValueTypeInteger },
-        { VASurfaceAttribNone, }
+        { VASurfaceAttribNone,          VAGenericValueTypeInteger }
     };
 
     if (!out_attribs || !out_num_attribs_ptr)
@@ -1105,6 +1142,7 @@ vaQuerySurfaceAttributes(
                                                          attrib_list, num_attribs);
 
     VA_TRACE_LOG(va_TraceQuerySurfaceAttributes, dpy, config, attrib_list, num_attribs);
+    VA_TRACE_RET(dpy, vaStatus); 
 
     return vaStatus;
 }
@@ -1141,6 +1179,7 @@ vaCreateSurfaces(
     VA_TRACE_LOG(va_TraceCreateSurfaces,
                  dpy, width, height, format, num_surfaces, surfaces,
                  attrib_list, num_attribs);
+    VA_TRACE_RET(dpy, vaStatus); 
 
     return vaStatus;
 }
@@ -1162,6 +1201,7 @@ VAStatus vaDestroySurfaces (
                dpy, surface_list, num_surfaces);
   
   vaStatus = ctx->vtable->vaDestroySurfaces( ctx, surface_list, num_surfaces );
+  VA_TRACE_RET(dpy, vaStatus); 
   
   return vaStatus;
 }
@@ -1188,6 +1228,7 @@ VAStatus vaCreateContext (
 
   /* keep current encode/decode resoluton */
   VA_TRACE_ALL(va_TraceCreateContext, dpy, config_id, picture_width, picture_height, flag, render_targets, num_render_targets, context);
+  VA_TRACE_RET(dpy, vaStatus); 
 
   return vaStatus;
 }
@@ -1206,6 +1247,7 @@ VAStatus vaDestroyContext (
   vaStatus = ctx->vtable->vaDestroyContext( ctx, context );
 
   VA_TRACE_ALL(va_TraceDestroyContext, dpy, context);
+  VA_TRACE_RET(dpy, vaStatus); 
 
   return vaStatus;
 }
@@ -1228,6 +1270,7 @@ VAStatus vaCreateMFContext (
         VA_TRACE_ALL(va_TraceCreateMFContext, dpy, mf_context);
     }
 
+    VA_TRACE_RET(dpy, vaStatus); 
     return vaStatus;
 }
 
@@ -1251,6 +1294,7 @@ VAStatus vaMFAddContext (
         VA_TRACE_ALL(va_TraceMFAddContext, dpy, context, mf_context);
     }
 
+    VA_TRACE_RET(dpy, vaStatus); 
     return vaStatus;
 }
 
@@ -1272,6 +1316,7 @@ VAStatus vaMFReleaseContext (
         vaStatus = ctx->vtable->vaMFReleaseContext( ctx, context, mf_context);
         VA_TRACE_ALL(va_TraceMFReleaseContext, dpy, context, mf_context);
     }
+    VA_TRACE_RET(dpy, vaStatus); 
 
     return vaStatus;
 }
@@ -1296,6 +1341,7 @@ VAStatus vaMFSubmit (
         vaStatus = ctx->vtable->vaMFSubmit( ctx, mf_context, contexts, num_contexts);
         VA_TRACE_ALL(va_TraceMFSubmit, dpy, mf_context, contexts, num_contexts);
     }
+    VA_TRACE_RET(dpy, vaStatus); 
 
     return vaStatus;
 }
@@ -1323,6 +1369,7 @@ VAStatus vaCreateBuffer (
   VA_TRACE_LOG(va_TraceCreateBuffer,
                dpy, context, type, size, num_elements, data, buf_id);
   
+  VA_TRACE_RET(dpy, vaStatus); 
   return vaStatus;
 }
 
@@ -1349,6 +1396,7 @@ VAStatus vaCreateBuffer2 (
 
   VA_TRACE_LOG(va_TraceCreateBuffer,
                dpy, context, type, *pitch, height, NULL, buf_id);
+  VA_TRACE_RET(dpy, vaStatus); 
 
   return vaStatus;
 }
@@ -1360,12 +1408,15 @@ VAStatus vaBufferSetNumElements (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
   VA_FOOL_FUNC(va_FoolCheckContinuity, dpy);
   
-  return ctx->vtable->vaBufferSetNumElements( ctx, buf_id, num_elements );
+  vaStatus = ctx->vtable->vaBufferSetNumElements( ctx, buf_id, num_elements );
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 
@@ -1386,6 +1437,7 @@ VAStatus vaMapBuffer (
   va_status = ctx->vtable->vaMapBuffer( ctx, buf_id, pbuf );
 
   VA_TRACE_ALL(va_TraceMapBuffer, dpy, buf_id, pbuf);
+  VA_TRACE_RET(dpy, va_status); 
   
   return va_status;
 }
@@ -1396,12 +1448,15 @@ VAStatus vaUnmapBuffer (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
   VA_FOOL_FUNC(va_FoolCheckContinuity, dpy);
 
-  return ctx->vtable->vaUnmapBuffer( ctx, buf_id );
+  vaStatus = ctx->vtable->vaUnmapBuffer( ctx, buf_id );
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 VAStatus vaDestroyBuffer (
@@ -1410,6 +1465,7 @@ VAStatus vaDestroyBuffer (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
@@ -1418,7 +1474,9 @@ VAStatus vaDestroyBuffer (
   VA_TRACE_LOG(va_TraceDestroyBuffer,
                dpy, buffer_id);
   
-  return ctx->vtable->vaDestroyBuffer( ctx, buffer_id );
+  vaStatus = ctx->vtable->vaDestroyBuffer( ctx, buffer_id );
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 VAStatus vaBufferInfo (
@@ -1431,13 +1489,16 @@ VAStatus vaBufferInfo (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
   
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
   VA_FOOL_FUNC(va_FoolBufferInfo, dpy, buf_id, type, size, num_elements);
   
-  return ctx->vtable->vaBufferInfo( ctx, buf_id, type, size, num_elements );
+  vaStatus = ctx->vtable->vaBufferInfo( ctx, buf_id, type, size, num_elements );
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus;
 }
 
 /* Locks buffer for external API usage */
@@ -1445,13 +1506,17 @@ VAStatus
 vaAcquireBufferHandle(VADisplay dpy, VABufferID buf_id, VABufferInfo *buf_info)
 {
     VADriverContextP ctx;
+    VAStatus vaStatus = VA_STATUS_SUCCESS;
 
     CHECK_DISPLAY(dpy);
     ctx = CTX(dpy);
 
     if (!ctx->vtable->vaAcquireBufferHandle)
-        return VA_STATUS_ERROR_UNIMPLEMENTED;
-    return ctx->vtable->vaAcquireBufferHandle(ctx, buf_id, buf_info);
+        vaStatus = VA_STATUS_ERROR_UNIMPLEMENTED;
+    else
+        vaStatus = ctx->vtable->vaAcquireBufferHandle(ctx, buf_id, buf_info);
+    VA_TRACE_RET(dpy, vaStatus);
+    return vaStatus;
 }
 
 /* Unlocks buffer after usage from external API */
@@ -1459,13 +1524,17 @@ VAStatus
 vaReleaseBufferHandle(VADisplay dpy, VABufferID buf_id)
 {
     VADriverContextP ctx;
+    VAStatus vaStatus = VA_STATUS_SUCCESS;
 
     CHECK_DISPLAY(dpy);
     ctx = CTX(dpy);
 
     if (!ctx->vtable->vaReleaseBufferHandle)
-        return VA_STATUS_ERROR_UNIMPLEMENTED;
-    return ctx->vtable->vaReleaseBufferHandle(ctx, buf_id);
+        vaStatus = VA_STATUS_ERROR_UNIMPLEMENTED;
+    else
+        vaStatus = ctx->vtable->vaReleaseBufferHandle(ctx, buf_id);
+    VA_TRACE_RET(dpy, vaStatus);
+    return vaStatus;
 }
 
 VAStatus
@@ -1474,15 +1543,19 @@ vaExportSurfaceHandle(VADisplay dpy, VASurfaceID surface_id,
                       void *descriptor)
 {
     VADriverContextP ctx;
+    VAStatus vaStatus = VA_STATUS_SUCCESS;
 
     CHECK_DISPLAY(dpy);
     ctx = CTX(dpy);
 
     if (!ctx->vtable->vaExportSurfaceHandle)
-        return VA_STATUS_ERROR_UNIMPLEMENTED;
-    return ctx->vtable->vaExportSurfaceHandle(ctx, surface_id,
+        vaStatus = VA_STATUS_ERROR_UNIMPLEMENTED;
+    else
+        vaStatus = ctx->vtable->vaExportSurfaceHandle(ctx, surface_id,
                                               mem_type, flags,
                                               descriptor);
+    VA_TRACE_RET(dpy, vaStatus);
+    return vaStatus;
 }
 
 VAStatus vaBeginPicture (
@@ -1501,6 +1574,7 @@ VAStatus vaBeginPicture (
   VA_FOOL_FUNC(va_FoolCheckContinuity, dpy);
   
   va_status = ctx->vtable->vaBeginPicture( ctx, context, render_target );
+  VA_TRACE_RET(dpy, va_status);
   
   return va_status;
 }
@@ -1513,6 +1587,7 @@ VAStatus vaRenderPicture (
 )
 {
   VADriverContextP ctx;
+  VAStatus vaStatus = VA_STATUS_SUCCESS;
 
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
@@ -1520,7 +1595,9 @@ VAStatus vaRenderPicture (
   VA_TRACE_LOG(va_TraceRenderPicture, dpy, context, buffers, num_buffers);
   VA_FOOL_FUNC(va_FoolCheckContinuity, dpy);
 
-  return ctx->vtable->vaRenderPicture( ctx, context, buffers, num_buffers );
+  vaStatus = ctx->vtable->vaRenderPicture( ctx, context, buffers, num_buffers );
+  VA_TRACE_RET(dpy, vaStatus);
+  return vaStatus; 
 }
 
 VAStatus vaEndPicture (
@@ -1535,11 +1612,11 @@ VAStatus vaEndPicture (
   ctx = CTX(dpy);
 
   VA_FOOL_FUNC(va_FoolCheckContinuity, dpy);
-
+  VA_TRACE_ALL(va_TraceEndPicture, dpy, context, 0);
   va_status = ctx->vtable->vaEndPicture( ctx, context );
-
+  VA_TRACE_RET(dpy, va_status);
   /* dump surface content */
-  VA_TRACE_ALL(va_TraceEndPicture, dpy, context, 1);
+  VA_TRACE_ALL(va_TraceEndPictureExt, dpy, context, 1);
 
   return va_status;
 }
@@ -1557,6 +1634,29 @@ VAStatus vaSyncSurface (
 
   va_status = ctx->vtable->vaSyncSurface( ctx, render_target );
   VA_TRACE_LOG(va_TraceSyncSurface, dpy, render_target);
+  VA_TRACE_RET(dpy, va_status);
+
+  return va_status;
+}
+
+VAStatus vaSyncSurface2 (
+    VADisplay dpy,
+    VASurfaceID surface,
+    uint64_t timeout_ns
+)
+{
+  VAStatus va_status;
+  VADriverContextP ctx;
+
+  CHECK_DISPLAY(dpy);
+  ctx = CTX(dpy);
+
+  if (ctx->vtable->vaSyncSurface2)
+      va_status = ctx->vtable->vaSyncSurface2( ctx, surface, timeout_ns );
+  else
+      va_status = VA_STATUS_ERROR_UNIMPLEMENTED;
+  VA_TRACE_LOG(va_TraceSyncSurface2, dpy, surface, timeout_ns);
+  VA_TRACE_RET(dpy, va_status);
 
   return va_status;
 }
@@ -1575,6 +1675,7 @@ VAStatus vaQuerySurfaceStatus (
   va_status = ctx->vtable->vaQuerySurfaceStatus( ctx, render_target, status );
 
   VA_TRACE_LOG(va_TraceQuerySurfaceStatus, dpy, render_target, status);
+  VA_TRACE_RET(dpy, va_status);
 
   return va_status;
 }
@@ -1594,6 +1695,30 @@ VAStatus vaQuerySurfaceError (
   va_status = ctx->vtable->vaQuerySurfaceError( ctx, surface, error_status, error_info );
 
   VA_TRACE_LOG(va_TraceQuerySurfaceError, dpy, surface, error_status, error_info);
+  VA_TRACE_RET(dpy, va_status); 
+
+  return va_status;
+}
+
+VAStatus vaSyncBuffer (
+    VADisplay dpy,
+    VABufferID buf_id,
+    uint64_t timeout_ns
+)
+{
+  VAStatus va_status;
+  VADriverContextP ctx;
+
+  CHECK_DISPLAY(dpy);
+  ctx = CTX(dpy);
+
+  VA_TRACE_LOG(va_TraceSyncBuffer, dpy, buf_id, timeout_ns);
+
+  if (ctx->vtable->vaSyncBuffer)
+      va_status = ctx->vtable->vaSyncBuffer( ctx, buf_id, timeout_ns );
+  else
+      va_status = VA_STATUS_ERROR_UNIMPLEMENTED;
+  VA_TRACE_RET(dpy, va_status);
 
   return va_status;
 }
@@ -1639,10 +1764,13 @@ VAStatus vaCreateImage (
 )
 {
   VADriverContextP ctx;
+  VAStatus va_status = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaCreateImage ( ctx, format, width, height, image);
+  va_status = ctx->vtable->vaCreateImage ( ctx, format, width, height, image);
+  VA_TRACE_RET(dpy, va_status);
+  return va_status;
 }
 
 /*
@@ -1654,10 +1782,13 @@ VAStatus vaDestroyImage (
 )
 {
   VADriverContextP ctx;
+  VAStatus va_status = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaDestroyImage ( ctx, image);
+  va_status = ctx->vtable->vaDestroyImage ( ctx, image);
+  VA_TRACE_RET(dpy, va_status);
+  return va_status;
 }
 
 VAStatus vaSetImagePalette (
@@ -1667,10 +1798,13 @@ VAStatus vaSetImagePalette (
 )
 {
   VADriverContextP ctx;
+  VAStatus va_status = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaSetImagePalette ( ctx, image, palette);
+  va_status = ctx->vtable->vaSetImagePalette ( ctx, image, palette);
+  VA_TRACE_RET(dpy, va_status);
+  return va_status;
 }
 
 /*
@@ -1688,10 +1822,13 @@ VAStatus vaGetImage (
 )
 {
   VADriverContextP ctx;
+  VAStatus va_status = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaGetImage ( ctx, surface, x, y, width, height, image);
+  va_status = ctx->vtable->vaGetImage ( ctx, surface, x, y, width, height, image);
+  VA_TRACE_RET(dpy, va_status);
+  return va_status;
 }
 
 /*
@@ -1713,10 +1850,13 @@ VAStatus vaPutImage (
 )
 {
   VADriverContextP ctx;
+  VAStatus va_status = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaPutImage ( ctx, surface, image, src_x, src_y, src_width, src_height, dest_x, dest_y, dest_width, dest_height );
+  va_status = ctx->vtable->vaPutImage ( ctx, surface, image, src_x, src_y, src_width, src_height, dest_x, dest_y, dest_width, dest_height );
+  VA_TRACE_RET(dpy, va_status);
+  return va_status;
 }
 
 /*
@@ -1757,10 +1897,13 @@ VAStatus vaDeriveImage (
 )
 {
   VADriverContextP ctx;
+  VAStatus va_status = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaDeriveImage ( ctx, surface, image );
+  va_status = ctx->vtable->vaDeriveImage ( ctx, surface, image );
+  VA_TRACE_RET(dpy, va_status);
+  return va_status;
 }
 
 
@@ -1969,6 +2112,7 @@ VAStatus vaQueryDisplayAttributes (
   va_status = ctx->vtable->vaQueryDisplayAttributes ( ctx, attr_list, num_attributes );
 
   VA_TRACE_LOG(va_TraceQueryDisplayAttributes, dpy, attr_list, num_attributes);
+  VA_TRACE_RET(dpy, va_status); 
 
   return va_status;
   
@@ -1994,6 +2138,7 @@ VAStatus vaGetDisplayAttributes (
   va_status = ctx->vtable->vaGetDisplayAttributes ( ctx, attr_list, num_attributes );
 
   VA_TRACE_LOG(va_TraceGetDisplayAttributes, dpy, attr_list, num_attributes);
+  VA_TRACE_RET(dpy, va_status); 
   
   return va_status;
 }
@@ -2017,6 +2162,7 @@ VAStatus vaSetDisplayAttributes (
 
   va_status = ctx->vtable->vaSetDisplayAttributes ( ctx, attr_list, num_attributes );
   VA_TRACE_LOG(va_TraceSetDisplayAttributes, dpy, attr_list, num_attributes);
+  VA_TRACE_RET(dpy, va_status); 
   
   return va_status;
 }
@@ -2035,10 +2181,14 @@ VAStatus vaLockSurface(VADisplay dpy,
 )
 {
   VADriverContextP ctx;
+  VAStatus va_status = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaLockSurface( ctx, surface, fourcc, luma_stride, chroma_u_stride, chroma_v_stride, luma_offset, chroma_u_offset, chroma_v_offset, buffer_name, buffer);
+  va_status = ctx->vtable->vaLockSurface( ctx, surface, fourcc, luma_stride, chroma_u_stride, chroma_v_stride, luma_offset, chroma_u_offset, chroma_v_offset, buffer_name, buffer);
+  VA_TRACE_RET(dpy, va_status);
+
+  return va_status;
 }
 
 
@@ -2047,10 +2197,14 @@ VAStatus vaUnlockSurface(VADisplay dpy,
 )
 {
   VADriverContextP ctx;
+  VAStatus va_status = VA_STATUS_SUCCESS;
   CHECK_DISPLAY(dpy);
   ctx = CTX(dpy);
 
-  return ctx->vtable->vaUnlockSurface( ctx, surface );
+  va_status = ctx->vtable->vaUnlockSurface( ctx, surface );
+  VA_TRACE_RET(dpy, va_status);
+
+  return va_status;
 }
 
 /* Video Processing */
@@ -2084,6 +2238,8 @@ vaQueryVideoProcFilters(
         QueryVideoProcFilters,
         (ctx, context, filters, num_filters)
     );
+    VA_TRACE_RET(dpy, status);
+
     return status;
 }
 
@@ -2105,6 +2261,7 @@ vaQueryVideoProcFilterCaps(
         QueryVideoProcFilterCaps,
         (ctx, context, type, filter_caps, num_filter_caps)
     );
+    VA_TRACE_RET(dpy, status);
     return status;
 }
 
@@ -2126,5 +2283,6 @@ vaQueryVideoProcPipelineCaps(
         QueryVideoProcPipelineCaps,
         (ctx, context, filters, num_filters, pipeline_caps)
     );
+    VA_TRACE_RET(dpy, status);
     return status;
 }
